@@ -5,11 +5,16 @@
 */
 
 #include "hafasmgateparser.h"
+#include "hafasvehiclelayoutparser.h"
 #include "logging.h"
+#include "datatypes/loadutil_p.h"
+#include "geo/polylinedecoder_p.h"
 
 #include <KPublicTransport/Journey>
 #include <KPublicTransport/Line>
+#include <KPublicTransport/Platform>
 #include <KPublicTransport/Stopover>
+#include <KPublicTransport/Vehicle>
 
 #include <QColor>
 #include <QDateTime>
@@ -17,13 +22,15 @@
 #include <QJsonArray>
 #include <QJsonDocument>
 #include <QJsonObject>
+#include <QRegularExpression>
 
 using namespace KPublicTransport;
 
 // REM or HIM elements
 struct Message {
-    QString text;
+    QVariant content;
     Disruption::Effect effect = Disruption::NormalService;
+    LoadInfo loadInfo;
 };
 
 HafasMgateParser::HafasMgateParser() = default;
@@ -49,12 +56,31 @@ static std::vector<Ico> parseIcos(const QJsonArray &icoL)
     return icos;
 }
 
+static constexpr const Load::Category load_value_map[] = {
+    Load::Unknown,
+    Load::Low, // 1
+    Load::Medium, // 2
+    Load::High, // 3
+    Load::Full // 4
+};
+
 static const struct {
     const char *type;
     const char *code;
 } ignored_remarks[] = {
+    { "A", "1" }, // different name formats for the line, used by SBB
+    { "A", "2" },
+    { "A", "3" },
+    { "A", "4" }, // same as above, containing product, line number and journey number
     { "A", "OPERATOR" }, // operator information should be a dedicated field if we ever need it
+    { "A", "moreAttr" }, // ZVV: pointless note about checking intermediate stops for more details
     { "H", "wagenstand_v2" }, // contains a pointless note about checking trip details
+    { "I", "FD" }, // SBB line number?
+    { "I", "RN" }, // SBB: some unknown number for buses
+    { "I", "TC" }, // SBB: some unknown number for buses
+    { "I", "XC" }, // SBB: some XML structure of unknown content, related to train/platform layouts
+    { "I", "XG" }, // SBB: some XML structure of unknown content, related to train/platform layouts
+    { "I", "XT" }, // SBB: some XML structure of unknown content, related to train/platform layouts
 };
 
 static std::vector<Message> parseRemarks(const QJsonArray &remL)
@@ -79,9 +105,27 @@ static std::vector<Message> parseRemarks(const QJsonArray &remL)
         }
 
         Message m;
-        m.text = remObj.value(QLatin1String("txtN")).toString();
-        if (code == QLatin1String("text.realtime.stop.cancelled")) {
-            m.effect = Disruption::NoService;
+        if (type == QLatin1Char('I') && code == QLatin1String("JF")) {
+            m.content = HafasVehicleLayoutParser::parseTrainFormation(remObj.value(QLatin1String("txtN")).toString().toUtf8());
+        } else if (type == QLatin1Char('I') && code == QLatin1String("XP")) {
+            m.content = HafasVehicleLayoutParser::parsePlatformSectors(remObj.value(QLatin1String("txtN")).toString().toUtf8());
+        } else if (type == QLatin1Char('A') && (code.startsWith(QLatin1String("text.occup.loc.")) || code.startsWith(QLatin1String("text.occup.jny.")))) {
+            static QRegularExpression rx(QStringLiteral("\\.(max|1st|2nd)\\.1([1-4])$"));
+            const auto match = rx.match(code);
+            if (match.hasMatch()) {
+                m.loadInfo.setLoad(load_value_map[match.captured(2).toInt()]);
+                if (match.captured(1) != QLatin1String("max")) {
+                    m.loadInfo.setSeatingClass(match.captured(1).left(1));
+                }
+            } else {
+                m.content = remObj.value(QLatin1String("txtN")).toString();
+            }
+        } else {
+            // generic text
+            m.content = remObj.value(QLatin1String("txtN")).toString();
+            if (code == QLatin1String("text.realtime.stop.cancelled") || code == QLatin1String("text.realtime.stop.entry.exit.disabled")) {
+                m.effect = Disruption::NoService;
+            }
         }
         rems.push_back(std::move(m));
     }
@@ -95,16 +139,16 @@ static std::vector<Message> parseWarnings(const QJsonArray &himL)
     for (const auto &himV : himL) {
         const auto himObj = himV.toObject();
         Message m;
-        m.text = himObj.value(QLatin1String("head")).toString() + QLatin1Char('\n')
+        m.content = QString(himObj.value(QLatin1String("head")).toString() + QLatin1Char('\n')
                + himObj.value(QLatin1String("lead")).toString() + QLatin1Char('\n')
-               + himObj.value(QLatin1String("text")).toString();
+               + himObj.value(QLatin1String("text")).toString());
         hims.push_back(m);
     }
     return hims;
 }
 
-template <typename T>
-static void parseMessageList(T &elem, const QJsonObject &obj, const std::vector<Message> &remarks, const std::vector<Message> &warnings)
+template <typename Func>
+static void processMessageList(const QJsonObject &obj, const std::vector<Message> &remarks, const std::vector<Message> &warnings, Func func)
 {
     const auto msgL = obj.value(QLatin1String("msgL")).toArray();
     QStringList notes;
@@ -127,21 +171,32 @@ static void parseMessageList(T &elem, const QJsonObject &obj, const std::vector<
             qCDebug(Log) << "Invalid message index:" << remX << msgType;
             continue;
         }
-        const auto msg = (*source)[remX];
-        elem.addNote(msg.text);
-        if (msg.effect == Disruption::NoService) {
-            elem.setDisruptionEffect(msg.effect);
-        }
+        const auto &msg = (*source)[remX];
+        func(msg, msgObj);
     }
 }
 
-static constexpr const Load::Category load_value_map[] = {
-    Load::Unknown,
-    Load::Low, // 1
-    Load::Medium, // 2
-    Load::High, // 3
-    Load::Full // 4
-};
+template <typename T>
+static void applyMessage(T &elem, const Message &msg)
+{
+    if (msg.content.type() == QVariant::String) {
+        elem.addNote(msg.content.toString());
+    }
+    if (msg.effect == Disruption::NoService) {
+        elem.setDisruptionEffect(msg.effect);
+    }
+    if (msg.loadInfo.load() != Load::Unknown) {
+        elem.setLoadInformation(LoadUtil::merge(elem.loadInformation(), {msg.loadInfo}));
+    }
+}
+
+template <typename T>
+static void parseMessageList(T &elem, const QJsonObject &obj, const std::vector<Message> &remarks, const std::vector<Message> &warnings)
+{
+    processMessageList(obj, remarks, warnings, [&elem](const Message &msg, const QJsonObject&) {
+        applyMessage(elem, msg);
+    });
+}
 
 static std::vector<LoadInfo> parseLoadInformation(const QJsonArray &tcocL)
 {
@@ -366,9 +421,20 @@ std::vector<Journey> HafasMgateParser::parseJourneys(const QByteArray &data) con
     return {};
 }
 
-static std::vector<LoadInfo> parseLoad(const QJsonObject &obj, const std::vector<LoadInfo> &loadInfos)
+static void setPlatformLayout(Stopover &stop, const Platform &platform) { stop.setPlatformLayout(platform); }
+static void setPlatformLayout(JourneySection &jny, const Platform &platform) { jny.setDeparturePlatformLayout(platform); }
+static void setVehicleLayout(Stopover &stop, const Vehicle &vehicle) { stop.setVehicleLayout(vehicle); }
+static void setVehicleLayout(JourneySection &jny, const Vehicle &vehicle) { jny.setDepartureVehicleLayout(vehicle); }
+
+template <typename T>
+static void parseTrainComposition(const QJsonObject &obj, T &result,
+                                  const std::vector<LoadInfo> &loadInfos,
+                                  const std::vector<Platform> &platforms,
+                                  const std::vector<Vehicle> &vehicles)
 {
     const auto dTrnCmpSX = obj.value(QLatin1String("dTrnCmpSX")).toObject();
+
+    // load
     const auto tcocX = dTrnCmpSX.value(QLatin1String("tcocX")).toArray();
     std::vector<LoadInfo> load;
     load.reserve(tcocX.size());
@@ -378,7 +444,91 @@ static std::vector<LoadInfo> parseLoad(const QJsonObject &obj, const std::vector
             load.push_back(loadInfos[i]);
         }
     }
-    return load;
+    result.setLoadInformation(LoadUtil::merge(std::move(load), result.loadInformation()));
+
+    // platform
+    const auto tcpdX = dTrnCmpSX.value(QLatin1String("tcpdX")).toInt(-1);
+    if (tcpdX >= 0 && tcpdX < (int)platforms.size()) {
+        setPlatformLayout(result, platforms[tcpdX]);
+    }
+
+    // vehicle
+    const auto stcGX = dTrnCmpSX.value(QLatin1String("stcGX")).toArray();
+    const auto vehicleIdx = stcGX.empty() ? -1 : stcGX.at(0).toInt(-1);
+    if (vehicleIdx >= 0 && vehicleIdx < (int)vehicles.size()) {
+        setVehicleLayout(result, vehicles[vehicleIdx]);
+    }
+}
+
+static std::vector<Path> parsePaths(const QJsonArray &polyL, const std::vector<Location> &locs)
+{
+    std::vector<Path> paths;
+    paths.reserve(polyL.size());
+    for (const auto &polyV : polyL) {
+        const auto polyObj = polyV.toObject();
+
+        // path coordinate index to location mapping
+        const auto ppLocRefL = polyObj.value(QLatin1String("ppLocRefL")).toArray();
+        // 2-dimensional differential encoded coordinates
+        const auto crdEncYX = polyObj.value(QLatin1String("crdEncYX")).toString().toUtf8();
+        PolylineDecoder<2> crdEncYXDecoder(crdEncYX.constData());
+        // crdEncDist: 1-dimensional integer values with differential encoding containing distances in meters
+        // crdEncZ: 1-dimensional, always 0?
+        // crdEncS: 1-dimensional, unknown meaning, but very low-entropy data
+        // crdEncF: 1-dimensional, always 0?
+
+        std::vector<PathSection> sections;
+        sections.reserve(ppLocRefL.size() - 1);
+        int prevPpIdx = 0;
+        QPointF prevCoord;
+        for (const auto ppLocRefV : ppLocRefL) {
+            const auto ppLocRef = ppLocRefV.toObject();
+            const auto ppIdx = ppLocRef.value(QLatin1String("ppIdx")).toInt();
+            if (ppIdx == 0 || ppIdx < prevPpIdx) {
+                continue;
+            }
+
+            QPolygonF poly;
+            poly.reserve(prevPpIdx - ppIdx + 2);
+            if (!prevCoord.isNull()) {
+                poly.push_back(prevCoord);
+            }
+            crdEncYXDecoder.readPolygon(poly, ppIdx - prevPpIdx + 1);
+            if (!poly.empty()) {
+                prevCoord = poly.back();
+            }
+
+            PathSection section;
+            section.setPath(std::move(poly));
+            prevPpIdx = ppIdx;
+
+            const auto locX = ppLocRef.value(QLatin1String("locX")).toInt();
+            if (locX >= 0 && locX < (int)locs.size()) {
+                section.setDescription(locs[locX].name());
+            }
+
+            sections.push_back(std::move(section));
+        }
+
+        Path path;
+        path.setSections(std::move(sections));
+        paths.push_back(std::move(path));
+    }
+    return paths;
+}
+
+static Path parsePolyG(const QJsonObject &obj, const std::vector<Path> &paths)
+{
+    const auto polyG = obj.value(QLatin1String("polyG")).toObject();
+    const auto polyXL = polyG.value(QLatin1String("polyXL")).toArray();
+    if (polyXL.size() != 1) {
+        return {};
+    }
+    const auto polyX = polyXL.at(0).toInt();
+    if (polyX >= 0 && polyX < (int)paths.size()) {
+        return paths[polyX];
+    }
+    return {};
 }
 
 std::vector<Journey> HafasMgateParser::parseTripSearch(const QJsonObject &obj) const
@@ -390,6 +540,9 @@ std::vector<Journey> HafasMgateParser::parseTripSearch(const QJsonObject &obj) c
     const auto remarks = parseRemarks(commonObj.value(QLatin1String("remL")).toArray());
     const auto warnings = parseWarnings(commonObj.value(QLatin1String("himL")).toArray());
     const auto loadInfos = parseLoadInformation(commonObj.value(QLatin1String("tcocL")).toArray());
+    const auto paths = parsePaths(commonObj.value(QLatin1String("polyL")).toArray(), locs);
+    const auto platforms = HafasVehicleLayoutParser::parsePlatforms(commonObj);
+    const auto vehicles = HafasVehicleLayoutParser::parseVehicleLayouts(commonObj);
 
     std::vector<Journey> res;
     const auto outConL = obj.value(QLatin1String("outConL")).toArray();
@@ -412,9 +565,9 @@ std::vector<Journey> HafasMgateParser::parseTripSearch(const QJsonObject &obj) c
             const auto dep = secObj.value(QLatin1String("dep")).toObject();
             section.setScheduledDepartureTime(parseDateTime(dateStr, dep.value(QLatin1String("dTimeS")), dep.value(QLatin1String("dTZOffset"))));
             section.setExpectedDepartureTime(parseDateTime(dateStr, dep.value(QLatin1String("dTimeR")), dep.value(QLatin1String("dTZOffset"))));
-            auto locIdx = dep.value(QLatin1String("locX")).toInt();
-            if ((unsigned int)locIdx < locs.size()) {
-                section.setFrom(locs[locIdx]);
+            const auto fromIdx = dep.value(QLatin1String("locX")).toInt(-1);
+            if ((unsigned int)fromIdx < locs.size()) {
+                section.setFrom(locs[fromIdx]);
             }
             section.setScheduledDeparturePlatform(dep.value(QLatin1String("dPlatfS")).toString());
             section.setExpectedDeparturePlatform(dep.value(QLatin1String("dPlatfR")).toString());
@@ -425,9 +578,9 @@ std::vector<Journey> HafasMgateParser::parseTripSearch(const QJsonObject &obj) c
             const auto arr = secObj.value(QLatin1String("arr")).toObject();
             section.setScheduledArrivalTime(parseDateTime(dateStr, arr.value(QLatin1String("aTimeS")), arr.value(QLatin1String("aTZOffset"))));
             section.setExpectedArrivalTime(parseDateTime(dateStr, arr.value(QLatin1String("aTimeR")), arr.value(QLatin1String("aTZOffset"))));
-            locIdx = arr.value(QLatin1String("locX")).toInt();
-            if ((unsigned int)locIdx < locs.size()) {
-                section.setTo(locs[locIdx]);
+            const auto toIdx = arr.value(QLatin1String("locX")).toInt(-1);
+            if ((unsigned int)toIdx < locs.size()) {
+                section.setTo(locs[toIdx]);
             }
             section.setScheduledArrivalPlatform(arr.value(QLatin1String("aPlatfS")).toString());
             section.setExpectedArrivalPlatform(arr.value(QLatin1String("aPlatfR")).toString());
@@ -458,7 +611,6 @@ std::vector<Journey> HafasMgateParser::parseTripSearch(const QJsonObject &obj) c
                     stops.reserve(stopL.size() - 2);
                     for (auto it = std::next(stopL.begin()); it != std::prev(stopL.end()); ++it) {
                         const auto stopObj = (*it).toObject();
-                        // TODO how does this look for individual stop skips during disruptions?
                         Stopover stop;
                         const auto locIdx = stopObj.value(QLatin1String("locX")).toInt();
                         if ((unsigned int)locIdx < locs.size()) {
@@ -470,20 +622,57 @@ std::vector<Journey> HafasMgateParser::parseTripSearch(const QJsonObject &obj) c
                         stop.setExpectedArrivalTime(parseDateTime(dateStr, stopObj.value(QLatin1String("aTimeR")), stopObj.value(QLatin1String("aTZOffset"))));
                         stop.setScheduledPlatform(stopObj.value(QLatin1String("dPlatfS")).toString());
                         stop.setExpectedPlatform(stopObj.value(QLatin1String("dPlatfR")).toString());
-                        stop.setLoadInformation(parseLoad(stopObj, loadInfos));
+                        if (stopObj.value(QLatin1String("aCncl")).toBool() || stopObj.value(QLatin1String("dCncl")).toBool()) {
+                            stop.setDisruptionEffect(Disruption::NoService);
+                        }
+                        parseMessageList(stop, stopObj, remarks, warnings);
+                        processMessageList(jnyObj, remarks, warnings, [&stop, locIdx](const Message &msg, const QJsonObject &msgObj) {
+                            const auto fromIdx = msgObj.value(QLatin1String("fLocX")).toInt(-1);
+                            const auto toIdx = msgObj.value(QLatin1String("tLocX")).toInt(-1);
+                            if (fromIdx != toIdx || fromIdx != locIdx) {
+                                return;
+                            }
+                            if (msg.content.userType() == qMetaTypeId<Platform>()) {
+                                stop.setPlatformLayout(Platform::merge(stop.platformLayout(), msg.content.value<Platform>()));
+                            } else if (msg.content.userType() == qMetaTypeId<Vehicle>()) {
+                                stop.setVehicleLayout(Vehicle::merge(stop.vehicleLayout(), msg.content.value<Vehicle>()));
+                            }
+                            applyMessage(stop, msg);
+                        });
+                        parseTrainComposition(stopObj, stop, loadInfos, platforms, vehicles);
                         stops.push_back(stop);
                     }
                     section.setIntermediateStops(std::move(stops));
                 }
 
-                parseMessageList(section, jnyObj, remarks, warnings);
-                section.setLoadInformation(parseLoad(dep, loadInfos));
-            } else if (typeStr == QLatin1String("WALK")) {
-                section.setMode(JourneySection::Walking);
-                section.setDistance(secObj.value(QLatin1String("gis")).toObject().value(QLatin1String("dist")).toInt());
-            } else if (typeStr == QLatin1String("TRSF")) {
-                section.setMode(JourneySection::Transfer);
-                section.setDistance(secObj.value(QLatin1String("gis")).toObject().value(QLatin1String("dist")).toInt());
+                processMessageList(jnyObj, remarks, warnings, [&section, fromIdx, toIdx](const Message &msg, const QJsonObject &msgObj) {
+                    const auto from = msgObj.value(QLatin1String("fLocX")).toInt(-1);
+                    const auto to = msgObj.value(QLatin1String("tLocX")).toInt(-1);
+                    if (from >= 0 && to >= 0 && from != fromIdx && toIdx != to && from == to) {
+                        return;
+                    }
+                    if (msg.content.userType() == qMetaTypeId<Platform>()) {
+                        if (from == to && to == toIdx) {
+                            section.setArrivalPlatformLayout(Platform::merge(section.arrivalPlatformLayout(), msg.content.value<Platform>()));
+                        } else {
+                            section.setDeparturePlatformLayout(Platform::merge(section.departurePlatformLayout(), msg.content.value<Platform>()));
+                        }
+                    } else if (msg.content.userType() == qMetaTypeId<Vehicle>()) {
+                        if (from == to && to == toIdx) {
+                            section.setArrivalVehicleLayout(Vehicle::merge(section.arrivalVehicleLayout(), msg.content.value<Vehicle>()));
+                        } else {
+                            section.setDepartureVehicleLayout(Vehicle::merge(section.departureVehicleLayout(), msg.content.value<Vehicle>()));
+                        }
+                    }
+                    applyMessage(section, msg);
+                });
+                parseTrainComposition(dep, section, loadInfos, platforms, vehicles);
+                section.setPath(parsePolyG(jnyObj, paths));
+            } else if (typeStr == QLatin1String("WALK") || typeStr == QLatin1String("TRSF")) {
+                const auto gis = secObj.value(QLatin1String("gis")).toObject();
+                section.setDistance(gis.value(QLatin1String("dist")).toInt());
+                section.setPath(parsePolyG(gis, paths));
+                section.setMode(typeStr == QLatin1String("WALK") ? JourneySection::Walking : JourneySection::Transfer);
             }
 
             sections.push_back(section);
