@@ -21,18 +21,21 @@
 #include "datatypes/disruption.h"
 #include "datatypes/platform.h"
 #include "datatypes/vehicle.h"
-#include "datatypes/vehiclelayoutresult_p.h"
+#include "geo/geojson_p.h"
 
 #include <KPublicTransport/Journey>
 #include <KPublicTransport/Location>
 #include <KPublicTransport/Stopover>
 
+#include "backends/accessibilitycloudbackend.h"
 #include "backends/cache.h"
 #include "backends/deutschebahnbackend.h"
 #include "backends/efabackend.h"
 #include "backends/hafasmgatebackend.h"
 #include "backends/hafasquerybackend.h"
+#include "backends/ivvassbackend.h"
 #include "backends/navitiabackend.h"
+#include "backends/oebbbackend.h"
 #include "backends/opentripplannergraphqlbackend.h"
 #include "backends/opentripplannerrestbackend.h"
 #include "gbfs/gbfsbackend.h"
@@ -64,9 +67,9 @@ public:
     void loadNetworks();
     std::unique_ptr<AbstractBackend> loadNetwork(const QJsonObject &obj);
     template <typename Backend, typename Backend2, typename ...Backends>
-    static std::unique_ptr<AbstractBackend> loadNetwork(const QString &backendType, const QJsonObject &obj);
+    static std::unique_ptr<AbstractBackend> loadNetwork(const QJsonObject &backendType, const QJsonObject &obj);
     template <typename Backend> std::unique_ptr<AbstractBackend>
-    static loadNetwork(const QString &backendType, const QJsonObject &obj);
+    static loadNetwork(const QJsonObject &backendType, const QJsonObject &obj);
     template <typename T>
     static std::unique_ptr<AbstractBackend> loadNetwork(const QJsonObject &obj);
 
@@ -178,32 +181,36 @@ void ManagerPrivate::loadNetworks()
 
 std::unique_ptr<AbstractBackend> ManagerPrivate::loadNetwork(const QJsonObject &obj)
 {
-    const auto type = obj.value(QLatin1String("type")).toString();
+    const auto type = obj.value(QLatin1String("type")).toObject();
+    // backends need to be topologically sorted according to their preference/priority here
     return loadNetwork<
         NavitiaBackend,
         OpenTripPlannerGraphQLBackend,
         OpenTripPlannerRestBackend,
+        DeutscheBahnBackend,
+        OebbBackend,
         HafasMgateBackend,
         HafasQueryBackend,
         EfaBackend,
-        DeutscheBahnBackend,
-        GBFSBackend
+        IvvAssBackend,
+        GBFSBackend,
+        AccessibilityCloudBackend
     >(type, obj);
 }
 
 template <typename Backend, typename Backend2, typename ...Backends>
-std::unique_ptr<AbstractBackend> ManagerPrivate::loadNetwork(const QString &backendType, const QJsonObject &obj)
+std::unique_ptr<AbstractBackend> ManagerPrivate::loadNetwork(const QJsonObject &backendType, const QJsonObject &obj)
 {
-    if (backendType == QLatin1String(Backend::type())) {
+    if (backendType.value(QLatin1String(Backend::type())).toBool()) {
         return loadNetwork<Backend>(obj);
     }
     return loadNetwork<Backend2, Backends...>(backendType, obj);
 }
 
 template <typename Backend>
-std::unique_ptr<AbstractBackend> ManagerPrivate::loadNetwork(const QString &backendType, const QJsonObject &obj)
+std::unique_ptr<AbstractBackend> ManagerPrivate::loadNetwork(const QJsonObject &backendType, const QJsonObject &obj)
 {
-    if (backendType == QLatin1String(Backend::type())) {
+    if (backendType.value(QLatin1String(Backend::type())).toBool()) {
         return ManagerPrivate::loadNetwork<Backend>(obj);
     }
     qCWarning(Log) << "Unknown backend type:" << backendType;
@@ -237,15 +244,13 @@ static void applyBackendOptions(AbstractBackend *backend, const QMetaObject *mo,
         }
     }
 
-    const auto filter = obj.value(QLatin1String("filter")).toObject();
-    const auto geoFilter = filter.value(QLatin1String("geo")).toArray();
-    if (!geoFilter.isEmpty()) {
-        QPolygonF poly;
-        poly.reserve(geoFilter.size());
-        for (const auto &coordV : geoFilter) {
-            const auto coordA = coordV.toArray();
-            poly.push_back({coordA[0].toDouble(), coordA[1].toDouble()});
+    const auto coverageData = obj.value(QLatin1String("coverage")).toObject();
+    for (const auto &coverageType : { "anyCoverage", "regularCoverage", "realtimeCoverage" }) {
+        const auto coverage = coverageData.value(QLatin1String(coverageType)).toObject();
+        if (coverage.empty()) {
+            continue;
         }
+        const auto poly = GeoJson::readOuterPolygon(coverage.value(QLatin1String("area")).toObject());
         backend->setGeoFilter(poly);
     }
 
@@ -713,7 +718,7 @@ VehicleLayoutReply* Manager::queryVehicleLayout(const VehicleLayoutRequest &req)
         if (d->shouldSkipBackend(backend.get(), req)) {
             continue;
         }
-        if (req.departure().stopPoint().hasCoordinate() && backend->isLocationExcluded(req.departure().stopPoint())) {
+        if (req.stopover().stopPoint().hasCoordinate() && backend->isLocationExcluded(req.stopover().stopPoint())) {
             qCDebug(Log) << "Skipping backend based on location filter:" << backend->backendId();
             continue;
         }
@@ -728,7 +733,7 @@ VehicleLayoutReply* Manager::queryVehicleLayout(const VehicleLayoutRequest &req)
                 qCDebug(Log) << "Positive cache hit for backend" << backend->backendId();
                 if (cache.data.size() == 1) {
                     reply->addAttributions(std::move(cache.attributions));
-                    reply->addResult(cache.data[0].vehicle(), cache.data[0].platform(), cache.data[0].departure());
+                    reply->addResult(cache.data[0]);
                     break;
                 }
                 Q_FALLTHROUGH();
